#include <stdio.h>
#include <glm/gtc/type_ptr.hpp>

#include "D3D11Renderer.hpp"

using Microsoft::WRL::ComPtr;

namespace
{
#define HCHECK(value) HCheck(#value, __LINE__, value)

void HCheck(const char* what, int line, HRESULT hr)
{
    if (FAILED(hr)) {
        printf("%s failed with code %d at line %d\n", what, hr, line);
        abort();
    }
}

ID3DBlob* compileShader(const char* src, const char* target, const char* name)
{
    ID3DBlob* compiledShader = nullptr;
    ID3DBlob* compilerMsgs = nullptr;
    HRESULT hr = D3DCompile(src, strlen(src) + 1, name, nullptr, nullptr, "main", target, D3DCOMPILE_OPTIMIZATION_LEVEL3, 0, &compiledShader, &compilerMsgs);

    if (compilerMsgs != nullptr) {
        printf("compiler errors: '%.*s'\n", static_cast<int>(compilerMsgs->GetBufferSize()), reinterpret_cast<char*>(compilerMsgs->GetBufferPointer()));
        compilerMsgs->Release();
    }

    HCHECK(hr);
    return compiledShader;
}

}  // namespace

ComPtr<IDXGIAdapter> getAdapter(varjo_Luid luid)
{
    ComPtr<IDXGIFactory> factory = nullptr;

    const HRESULT hr = CreateDXGIFactory1(IID_PPV_ARGS(&factory));
    if (SUCCEEDED(hr)) {
        UINT i = 0;
        while (true) {
            ComPtr<IDXGIAdapter> adapter = nullptr;
            if (factory->EnumAdapters(i++, &adapter) == DXGI_ERROR_NOT_FOUND) break;
            DXGI_ADAPTER_DESC desc;
            if (SUCCEEDED(adapter->GetDesc(&desc)) && desc.AdapterLuid.HighPart == luid.high && desc.AdapterLuid.LowPart == luid.low) {
                return adapter;
            }
        }
    }
    return nullptr;
}

D3D11Renderer::D3D11Renderer(varjo_Session* session, bool useMultiTexture)
    : IRenderer(session, useMultiTexture)
{
#ifdef _DEBUG
    UINT flags = D3D11_CREATE_DEVICE_DEBUG;
#else
    UINT flags = 0;
#endif
    ComPtr<IDXGIAdapter> adapter = getAdapter(varjo_D3D11GetLuid(session));
    HRESULT result = D3D11CreateDevice(adapter.Get(), adapter ? D3D_DRIVER_TYPE_UNKNOWN : D3D_DRIVER_TYPE_HARDWARE, 0, flags, nullptr, 0, D3D11_SDK_VERSION,
        &m_device, nullptr, &m_deviceContext);
    if (result != S_OK) {
        printf("Failed to create device: %s", getLastErrorString());
        abort();
    }

    createShaders();
    createPerFrameBuffers();

    createDepthStencilStates();
    createRasterizerState();

    m_deviceContext->IASetPrimitiveTopology(D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST);
}

D3D11Renderer::~D3D11Renderer()
{
    // Shut down the Varjo graphics.
    // This frees all of the swap chain textures and the graphics info structure.
    varjo_D3D11ShutDown(m_session);

    for (auto& buffer : m_perFrameBuffers) {
        buffer.constantBuffer->Release();
        buffer.instanceBuffer->Release();
    }

    m_depthStencilState->Release();
    m_gridDepthStencilState->Release();

    m_rasterizerState->Release();
    m_deviceContext->Release();
    m_device->Release();
}

std::shared_ptr<Geometry> D3D11Renderer::createGeometry(uint32_t vertexCount, uint32_t indexCount)
{
    return std::make_shared<D3D11Geometry>(this, vertexCount, indexCount);
}

std::shared_ptr<RenderTarget> D3D11Renderer::createRenderTarget(int32_t width, int32_t height, varjo_TextureFormat format)
{
    return std::make_shared<D3D11RenderTarget>(this, width, height, format);
}

std::shared_ptr<RenderTarget> D3D11Renderer::createRenderTarget(int32_t width, int32_t height, varjo_Texture texture)
{
    return std::make_shared<D3D11RenderTarget>(this, width, height, varjo_ToD3D11Texture(texture));
}

bool D3D11Renderer::initVarjo()
{
    // Initialize Varjo graphics API to use D3D11.
    // Retuns a pointer to varjo_GraphicsInfo which contains swap chain textures and their sizes.
    m_graphicsInfo = varjo_D3D11Init(m_session, m_device, varjo_TextureFormat_R8G8B8A8_SRGB, nullptr);

    // Check if the initialization was successful.
    const varjo_Error error = varjo_GetError(m_session);
    if (error != varjo_NoError) {
        printf("Failed to create Dx swap chain: %s", varjo_GetErrorDesc(error));
        return false;
    }
    return true;
}

void D3D11Renderer::useRenderTarget(const std::shared_ptr<RenderTarget>& renderTarget)
{
    auto dxRenderTarget = std::static_pointer_cast<D3D11RenderTarget, RenderTarget>(renderTarget);

    ID3D11RenderTargetView* tgts[]{dxRenderTarget->renderTargetView()};
    m_deviceContext->OMSetRenderTargets(1, tgts, dxRenderTarget->depthStencilView());
}

void D3D11Renderer::clearRenderTarget(const std::shared_ptr<RenderTarget>& renderTarget, float r, float g, float b, float a)
{
    auto dxRenderTarget = std::static_pointer_cast<D3D11RenderTarget, RenderTarget>(renderTarget);
    useRenderTarget(renderTarget);

    const float color[]{r, g, b, a};
    m_deviceContext->ClearRenderTargetView(dxRenderTarget->renderTargetView(), color);
    m_deviceContext->ClearDepthStencilView(dxRenderTarget->depthStencilView(), D3D11_CLEAR_DEPTH | D3D11_CLEAR_STENCIL, 1.0f, 255);
}

void D3D11Renderer::useGeometry(const std::shared_ptr<Geometry>& geometry)
{
    auto dxGeometry = std::static_pointer_cast<D3D11Geometry, Geometry>(geometry);

    ID3D11Buffer* vertexBuffer = dxGeometry->vertexBuffer();

    uint32_t stride = sizeof(Geometry::Vertex);
    uint32_t offset = 0;

    m_deviceContext->IASetVertexBuffers(0, 1, &vertexBuffer, &stride, &offset);
    m_deviceContext->IASetIndexBuffer(dxGeometry->indexBuffer(), DXGI_FORMAT_R32_UINT, 0);

    m_currentGeometry = geometry;
}

void D3D11Renderer::setupCamera(const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix)
{
    m_shaderConstants.viewMatrix = glm::transpose(viewMatrix);
    m_shaderConstants.projectionMatrix = glm::transpose(projectionMatrix);

    // Update constants.
    m_deviceContext->UpdateSubresource(m_perFrameBuffers[m_currentFrameBuffer].constantBuffer, 0, nullptr, &m_shaderConstants, 0, 0);
    m_deviceContext->VSSetConstantBuffers(0, 1, &m_perFrameBuffers[m_currentFrameBuffer].constantBuffer);

    unsigned int stride = sizeof(glm::mat4);
    unsigned int offset = 0;
    m_deviceContext->IASetVertexBuffers(1, 1, &m_perFrameBuffers[m_currentFrameBuffer].instanceBuffer, &stride, &offset);
}

void D3D11Renderer::setViewport(int32_t x, int32_t y, int32_t width, int32_t height)
{
    D3D11_VIEWPORT viewport{};

    viewport.TopLeftX = static_cast<float>(x);
    viewport.TopLeftY = static_cast<float>(y);
    viewport.Width = static_cast<float>(width);
    viewport.Height = static_cast<float>(height);
    viewport.MinDepth = 0;
    viewport.MaxDepth = 1;

    m_deviceContext->RSSetViewports(1, &viewport);
}

void D3D11Renderer::drawGrid()
{
    m_deviceContext->OMSetDepthStencilState(m_gridDepthStencilState, 0);

    m_deviceContext->VSSetShader(m_gridShader.vertexShader, nullptr, 0);
    m_deviceContext->IASetInputLayout(m_gridShader.inputLayout);
    m_deviceContext->PSSetShader(m_gridShader.pixelShader, nullptr, 0);
    m_deviceContext->DrawIndexed(m_currentGeometry->indexCount(), 0, 0);

    m_deviceContext->OMSetDepthStencilState(m_depthStencilState, 0);
}

void D3D11Renderer::queueObjects(const std::vector<glm::mat4>& matrices)
{
    if (m_instanceBuffer.queued > 0) {
        dispatchDraws();
    }

    // Map the instance buffer.
    D3D11_MAPPED_SUBRESOURCE mapped{};
    const HRESULT result = m_deviceContext->Map(m_perFrameBuffers[m_currentFrameBuffer].instanceBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mapped);
    if (result != S_OK) {
        printf("Failed to map instance buffer: %s, %d", getLastErrorString(), GetLastError());
        exit(1);
    }

    m_instanceBuffer.data = reinterpret_cast<glm::mat4*>(mapped.pData);
    memcpy(m_instanceBuffer.data, matrices.data(), matrices.size() * sizeof(glm::mat4));

    m_instanceBuffer.queued = static_cast<int32_t>(matrices.size());

    m_deviceContext->Unmap(m_perFrameBuffers[m_currentFrameBuffer].instanceBuffer, 0);
}

void D3D11Renderer::dispatchDraws()
{
    m_deviceContext->VSSetShader(m_defaultShader.vertexShader, nullptr, 0);
    m_deviceContext->IASetInputLayout(m_defaultShader.inputLayout);
    m_deviceContext->PSSetShader(m_defaultShader.pixelShader, nullptr, 0);
    m_deviceContext->DrawIndexedInstanced(m_currentGeometry->indexCount(), m_instanceBuffer.queued, 0, 0, 0);
    m_instanceBuffer.queued = 0;
}

void D3D11Renderer::advance() { m_currentFrameBuffer = (m_currentFrameBuffer + 1) % m_perFrameBuffers.size(); }

void D3D11Renderer::createShaders()
{
    createShader();
    createGridShader();
}

void D3D11Renderer::createShader()
{
    ID3DBlob* compiledVertexShader = compileShader(
        "cbuffer ConstantBuffer : register(b0) {\n"
        "  matrix view;\n"
        "  matrix projection;\n"
        "};\n"
        "struct VsInput {\n"
        "  float3 pos : POSITION;\n"
        "  float3 normal : NORMAL;\n"
        "  float4 world0 : TEXCOORD0;\n"
        "  float4 world1 : TEXCOORD1;\n"
        "  float4 world2 : TEXCOORD2;\n"
        "  float4 world3 : TEXCOORD3;\n"
        "};\n"
        "struct VsOutput {\n"
        "  float4 position : SV_POSITION;\n"
        "  float3 normal : TEXCOORD0;\n"
        "};\n"
        "VsOutput main(VsInput input) {\n"
        "  VsOutput output;\n"
        "\n"
        "  matrix world = matrix(input.world0, input.world1, input.world2, input.world3);\n"
        "\n"
        "  float4 pos = float4(input.pos, 1.0f);\n"
        "  pos = mul(pos, world);\n"
        "  pos = mul(pos, view);\n"
        "  pos = mul(pos, projection);\n"
        "\n"
        "  output.position = pos;\n"
        "  output.normal = mul(float4(input.normal, 0.0f), world).xyz;\n"
        "  return output;\n"
        "}\n",
        "vs_4_0", "vertex");

    ID3DBlob* compiledPixelShader = compileShader(
        "struct PsInput {\n"
        "  float4 position : SV_POSITION;\n"
        "  float3 normal : TEXCOORD0;\n"
        "};\n"
        "float4 main(PsInput input) : SV_TARGET {\n"
        "  return float4(input.normal, 1);\n"
        "}\n",
        "ps_4_0", "pixel");

    HRESULT result =
        m_device->CreateVertexShader(compiledVertexShader->GetBufferPointer(), compiledVertexShader->GetBufferSize(), nullptr, &m_defaultShader.vertexShader);
    if (result != S_OK) {
        printf("Failed to compile vertex shader: %d", GetLastError());
        abort();
    }

    result = m_device->CreatePixelShader(compiledPixelShader->GetBufferPointer(), compiledPixelShader->GetBufferSize(), nullptr, &m_defaultShader.pixelShader);
    if (result != S_OK) {
        printf("Failed to compile pixel shader: %d", GetLastError());
        abort();
    }

    D3D11_INPUT_ELEMENT_DESC inputElements[6] = {
        {"POSITION", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 0, D3D11_INPUT_PER_VERTEX_DATA, 0},
        {"NORMAL", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 12, D3D11_INPUT_PER_VERTEX_DATA, 0},
        {"TEXCOORD", 0, DXGI_FORMAT_R32G32B32A32_FLOAT, 1, 0, D3D11_INPUT_PER_INSTANCE_DATA, 1},
        {"TEXCOORD", 1, DXGI_FORMAT_R32G32B32A32_FLOAT, 1, 16, D3D11_INPUT_PER_INSTANCE_DATA, 1},
        {"TEXCOORD", 2, DXGI_FORMAT_R32G32B32A32_FLOAT, 1, 32, D3D11_INPUT_PER_INSTANCE_DATA, 1},
        {"TEXCOORD", 3, DXGI_FORMAT_R32G32B32A32_FLOAT, 1, 48, D3D11_INPUT_PER_INSTANCE_DATA, 1},
    };

    result = m_device->CreateInputLayout(
        inputElements, 6, compiledVertexShader->GetBufferPointer(), compiledVertexShader->GetBufferSize(), &m_defaultShader.inputLayout);
    if (result != S_OK) {
        printf("Failed to create input layout: %d", GetLastError());
        abort();
    }
}

void D3D11Renderer::createGridShader()
{
    ID3DBlob* compiledVertexShader = compileShader(
        "cbuffer ConstantBuffer : register(b0) {\n"
        "  matrix view;\n"
        "  matrix projection;\n"
        "};\n"
        "struct VsInput {\n"
        "  float3 pos : POSITION;\n"
        "  float3 normal : NORMAL;\n"
        "};\n"
        "struct VsOutput {\n"
        "  float4 position : SV_POSITION;\n"
        "  float3 vPosition: TEXCOORD0;\n"
        "};\n"
        "VsOutput main(VsInput input) {\n"
        "  VsOutput output;\n"
        "\n"
        "  matrix viewMatrix = view;\n"
        "  viewMatrix[3][0] = 0.0;\n"
        "  viewMatrix[3][1] = 0.0;\n"
        "  viewMatrix[3][2] = 0.0;\n"
        "\n"
        "  float4 pos = float4(input.pos, 1.0f);\n"
        "  pos = mul(pos, viewMatrix);\n"
        "  pos = mul(pos, projection);\n"
        "\n"
        "  output.position = pos;\n"
        "  output.vPosition = input.pos.xyz + 0.5;\n"
        "  return output;\n"
        "}\n",
        "vs_4_0", "vertex");

    ID3DBlob* compiledPixelShader = compileShader(
        "float grid(float coordinate, float size) {\n"
        "    float inRange = step(0.001, 1.0 - coordinate) * step(0.001, coordinate);\n"
        "    float x = (coordinate * size);\n"
        "    return inRange * step(0.25, x - floor(x)) * step(0.25, 1.0 - (x - floor(x)));\n"
        "}\n"
        "\n"
        "struct PsInput {\n"
        "  float4 position : SV_POSITION;\n"
        "  float3 vPosition : TEXCOORD0;\n"
        "};\n"
        "\n"
        "float4 main(PsInput input) : SV_TARGET {\n"
        "  float x = grid(input.vPosition.x, 4.0);\n"
        "  float y = grid(input.vPosition.y, 4.0);\n"
        "  float z = grid(input.vPosition.z, 4.0);\n"
        "  float value = saturate(x + y + z);\n"
        "\n"
        "  float3 color = step(0.999, input.vPosition.z) * (float3)lerp(0.65, 0.427, value);\n"
        "  color += step(0.999, 1.0 - input.vPosition.z) * (float3)lerp(1.0, 0.0, value);\n"
        "\n"
        "  color += step(0.999, input.vPosition.x) * lerp(float3(1, 0, 0), float3(0, 1, 0), value);\n"
        "  color += step(0.999, 1.0 - input.vPosition.x) * lerp(float3(1, 0, 1), float3(0, 0, 1), value);\n"
        "\n"
        "  color += step(0.999, input.vPosition.y) * lerp(float3(1, 1, 0), float3(0, 1, 1), value);\n"
        "  color += step(0.999, 1.0 - input.vPosition.y) * lerp(float3(0.25, 0, 0.392), float3(0, 0.392, 0.129), value);\n"
        "\n"
        "  return float4(color, 1.0);\n"
        "}\n",
        "ps_4_0", "pixel");

    HRESULT result =
        m_device->CreateVertexShader(compiledVertexShader->GetBufferPointer(), compiledVertexShader->GetBufferSize(), nullptr, &m_gridShader.vertexShader);
    if (result != S_OK) {
        printf("Failed to compile vertex shader: %d", GetLastError());
        abort();
    }

    result = m_device->CreatePixelShader(compiledPixelShader->GetBufferPointer(), compiledPixelShader->GetBufferSize(), nullptr, &m_gridShader.pixelShader);
    if (result != S_OK) {
        printf("Failed to compile pixel shader: %d", GetLastError());
        abort();
    }

    D3D11_INPUT_ELEMENT_DESC inputElements[2] = {
        {"POSITION", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 0, D3D11_INPUT_PER_VERTEX_DATA, 0},
        {"NORMAL", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 12, D3D11_INPUT_PER_VERTEX_DATA, 0},
    };

    result = m_device->CreateInputLayout(
        inputElements, 2, compiledVertexShader->GetBufferPointer(), compiledVertexShader->GetBufferSize(), &m_gridShader.inputLayout);
    if (result != S_OK) {
        printf("Failed to create input layout: %d", GetLastError());
        abort();
    }
}

void D3D11Renderer::createDepthStencilStates()
{
    D3D11_DEPTH_STENCIL_DESC depthStencilDesc{};
    depthStencilDesc.StencilEnable = false;

    depthStencilDesc.DepthEnable = true;
    depthStencilDesc.DepthWriteMask = D3D11_DEPTH_WRITE_MASK_ALL;
    depthStencilDesc.DepthFunc = D3D11_COMPARISON_LESS;

    HRESULT result = m_device->CreateDepthStencilState(&depthStencilDesc, &m_depthStencilState);
    if (result != S_OK) {
        printf("Failed to create depth stencil state: %d", result);
        abort();
    }

    depthStencilDesc.DepthEnable = true;
    depthStencilDesc.DepthWriteMask = D3D11_DEPTH_WRITE_MASK_ZERO;
    depthStencilDesc.DepthFunc = D3D11_COMPARISON_ALWAYS;

    result = m_device->CreateDepthStencilState(&depthStencilDesc, &m_gridDepthStencilState);
    if (result != S_OK) {
        printf("Failed to create grid depth stencil state: %d", result);
        abort();
    }
}

void D3D11Renderer::createRasterizerState()
{
    D3D11_RASTERIZER_DESC rasterizerDesc{};
    rasterizerDesc.FillMode = D3D11_FILL_SOLID;
    rasterizerDesc.CullMode = D3D11_CULL_BACK;
    rasterizerDesc.FrontCounterClockwise = true;
    rasterizerDesc.DepthBias = 0;
    rasterizerDesc.DepthBiasClamp = 0;
    rasterizerDesc.SlopeScaledDepthBias = 1.0f;
    rasterizerDesc.DepthClipEnable = true;
    rasterizerDesc.ScissorEnable = false;
    rasterizerDesc.MultisampleEnable = false;
    rasterizerDesc.AntialiasedLineEnable = false;

    HRESULT result = m_device->CreateRasterizerState(&rasterizerDesc, &m_rasterizerState);
    if (result != S_OK) {
        printf("Failed to create rasterizer state: %d", GetLastError());
        abort();
    }

    m_deviceContext->RSSetState(m_rasterizerState);
}

void D3D11Renderer::createPerFrameBuffers()
{
    for (int i = 0; i < 4 * 4; ++i) {
        auto instanceBuffer = createInstanceBuffer();
        auto constantBuffer = createConstantBuffer();

        m_perFrameBuffers.push_back({constantBuffer, instanceBuffer});
    }
}

ID3D11Buffer* D3D11Renderer::createInstanceBuffer()
{
    const int32_t maxInstances = 5000;
    m_instanceBuffer.maxInstances = maxInstances;

    D3D11_BUFFER_DESC bufferDesc{};
    bufferDesc.Usage = D3D11_USAGE_DYNAMIC;
    bufferDesc.BindFlags = D3D11_BIND_VERTEX_BUFFER;
    bufferDesc.ByteWidth = sizeof(glm::mat4) * maxInstances;
    bufferDesc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;
    bufferDesc.StructureByteStride = 0;

    ID3D11Buffer* buffer = nullptr;
    HRESULT result = m_device->CreateBuffer(&bufferDesc, nullptr, &buffer);
    if (result != S_OK) {
        printf("Failed to create constant buffer: %d", GetLastError());
        abort();
    }

    return buffer;
}

ID3D11Buffer* D3D11Renderer::createConstantBuffer()
{
    D3D11_BUFFER_DESC bufferDesc{};
    bufferDesc.Usage = D3D11_USAGE_DEFAULT;
    bufferDesc.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
    bufferDesc.ByteWidth = sizeof m_shaderConstants;
    bufferDesc.CPUAccessFlags = 0;
    bufferDesc.StructureByteStride = 0;

    ID3D11Buffer* buffer = nullptr;
    HRESULT result = m_device->CreateBuffer(&bufferDesc, nullptr, &buffer);
    if (result != S_OK) {
        printf("Failed to create constant buffer: %d", GetLastError());
        abort();
    }
    return buffer;
}
