#pragma once

#include <vector>
#include <DirectXMath.h>
#include <d3d11_1.h>
#include <d3dcompiler.h>
#include <windows.h>
#include <wrl.h>
#include <Varjo_d3d11.h>

#include "IRenderer.hpp"

class D3D11Renderer : public IRenderer
{
public:
    D3D11Renderer(varjo_Session* session, bool useMultiTexture);
    ~D3D11Renderer();

    ID3D11Device* dxDevice() const { return m_device; }
    ID3D11DeviceContext* dxDeviceContext() const { return m_deviceContext; }

    std::shared_ptr<Geometry> createGeometry(uint32_t vertexCount, uint32_t indexCount) override;
    std::shared_ptr<RenderTarget> createRenderTarget(int32_t width, int32_t height, varjo_TextureFormat format) override;
    std::shared_ptr<RenderTarget> createRenderTarget(int32_t width, int32_t height, varjo_Texture texture) override;

private:
    bool initVarjo() override;

    void useRenderTarget(const std::shared_ptr<RenderTarget>& renderTarget) override;
    void clearRenderTarget(const std::shared_ptr<RenderTarget>& renderTarget, float r, float g, float b, float a) override;

    void useGeometry(const std::shared_ptr<Geometry>& geometry) override;

    void setupCamera(const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix) override;
    void setViewport(int32_t x, int32_t y, int32_t width, int32_t height) override;

    void drawGrid() override;
    void queueObjects(const std::vector<glm::mat4>& matrices) override;
    void dispatchDraws() override;
    void advance() override;

    void createShaders();
    void createShader();
    void createGridShader();

    void createDepthStencilStates();
    void createRasterizerState();

    void createPerFrameBuffers();
    ID3D11Buffer* createInstanceBuffer();
    ID3D11Buffer* createConstantBuffer();

    struct ShaderConstants {
        glm::mat4 viewMatrix;
        glm::mat4 projectionMatrix;
    };

    struct InstanceBuffer {
        int32_t maxInstances;
        int32_t queued;
        glm::mat4* data;
    };

    struct PerFrameBuffers {
        ID3D11Buffer* constantBuffer;
        ID3D11Buffer* instanceBuffer;
    };

    struct Shader {
        ~Shader()
        {
            vertexShader->Release();
            pixelShader->Release();
            inputLayout->Release();
        }

        ID3D11VertexShader* vertexShader;
        ID3D11PixelShader* pixelShader;
        ID3D11InputLayout* inputLayout;
    };

    ShaderConstants m_shaderConstants;
    InstanceBuffer m_instanceBuffer = {};

    Shader m_defaultShader = {};
    Shader m_gridShader = {};

    int32_t m_currentFrameBuffer = 0;
    std::vector<PerFrameBuffers> m_perFrameBuffers;

    ID3D11DepthStencilState* m_depthStencilState = nullptr;
    ID3D11DepthStencilState* m_gridDepthStencilState = nullptr;

    ID3D11RasterizerState* m_rasterizerState = nullptr;
    ID3D11DeviceContext* m_deviceContext = nullptr;
    ID3D11Device* m_device = nullptr;
};
