#include <GL/glew.h>
#include <GL/wglew.h>
#include <glm/gtc/type_ptr.hpp>

#include "GLRenderer.hpp"

namespace
{
GLuint compileShader(GLenum type, const char* source, const char* name);
GLuint linkProgram(GLuint vertexShader, GLuint fragmentShader, const char* name);

LRESULT CALLBACK wndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
}  // namespace

GLRenderer::GLRenderer(varjo_Session* session, bool useMultiTexture)
    : IRenderer(session, useMultiTexture)
{
    createDummyWindow();

    if (wglewInit() != GLEW_OK) {
        printf("Failed to initialize WGLEW\n");
        return;
    }

    m_hdc = GetDC(m_hwnd);
    if (m_hdc == NULL) {
        printf("Failed to get DC: %s\n", getLastErrorString());
        return;
    }

    PIXELFORMATDESCRIPTOR pfd = {sizeof(PIXELFORMATDESCRIPTOR), 1,
        PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,  //
        PFD_TYPE_RGBA,                                               // RGBA colors.
        32,                                                          // 32 bit framebuffer
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        24,  // Number of bits for the depthbuffer
        8,   // Number of bits for the stencilbuffer
        0,   // Number of Aux buffers in the framebuffer.
        PFD_MAIN_PLANE, 0, 0, 0, 0};

    int pixelFormat = ChoosePixelFormat(m_hdc, &pfd);

    if (SetPixelFormat(m_hdc, pixelFormat, &pfd) == false) {
        printf("Failed to set pixel format: %s", getLastErrorString());
        return;
    }

    m_hglrc = wglCreateContext(m_hdc);
    if (m_hglrc == NULL) {
        printf("Failed to create OpenGL context: %s\n", getLastErrorString());
        return;
    }

    if (wglMakeCurrent(m_hdc, m_hglrc) == false) {
        printf("Failed to set current OpenGL context: %s\n", getLastErrorString());
        return;
    }

    if (glewInit() != GLEW_OK) {
        printf("Failed to initialize GLEW\n");
        return;
    }

    compileShaders();
    createPerFrameBuffers();

    // glEnable(GL_FRAMEBUFFER_SRGB); <- Doesn't work when copying textures
    glEnable(GL_SCISSOR_TEST);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);
    glDepthMask(GL_TRUE);
    glDepthFunc(GL_LEQUAL);
}

GLRenderer::~GLRenderer()
{
    // Shut down the Varjo graphics.
    // This frees all of the swap chain textures and the graphics info structure.
    varjo_GLShutDown(m_session);

    glDeleteProgram(m_program);
    glDeleteProgram(m_gridProgram);

    for (auto& buffer : m_perFrameBuffers) {
        glDeleteBuffers(1, &buffer.instanceBuffer);
        glDeleteBuffers(1, &buffer.uniformBuffer);
    }

    wglDeleteContext(m_hglrc);
    ReleaseDC(m_hwnd, m_hdc);
    DestroyWindow(m_hwnd);
}

std::shared_ptr<Geometry> GLRenderer::createGeometry(uint32_t vertexCount, uint32_t indexCount)
{
    return std::make_shared<GLGeometry>(vertexCount, indexCount);
}

std::shared_ptr<RenderTarget> GLRenderer::createRenderTarget(int32_t width, int32_t height, varjo_TextureFormat format)
{
    return std::make_shared<GLRenderTarget>(width, height, format);
}

std::shared_ptr<RenderTarget> GLRenderer::createRenderTarget(int32_t width, int32_t height, varjo_Texture texture)
{
    return std::make_shared<GLRenderTarget>(width, height, varjo_ToGLTexture(texture));
}

bool GLRenderer::initVarjo()
{
    // Initialize Varjo graphics API to use OpenGL.
    // Retuns a pointer to varjo_GraphicsInfo which contains swap chain textures and their sizes.
    m_graphicsInfo = varjo_GLInit(m_session, varjo_TextureFormat_R8G8B8A8_SRGB, nullptr);

    // Check if the initialization was successful.
    const varjo_Error error = varjo_GetError(m_session);
    if (error != varjo_NoError) {
        printf("Failed to create GL swap chain: %s", varjo_GetErrorDesc(error));
        return false;
    }
    return true;
}

void GLRenderer::useRenderTarget(const std::shared_ptr<RenderTarget>& renderTarget)
{
    auto glRenderTarget = std::static_pointer_cast<GLRenderTarget, RenderTarget>(renderTarget);

    glBindFramebuffer(GL_FRAMEBUFFER, glRenderTarget->frameBuffer());

    GLenum error = glGetError();
    if (error != GL_NO_ERROR) {
        printf("Failed to bind framebuffer: %x\n", error);
        abort();
    }
}

void GLRenderer::clearRenderTarget(const std::shared_ptr<RenderTarget>& renderTarget, float r, float g, float b, float a)
{
    useRenderTarget(renderTarget);

    glClearColor(r, g, b, a);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    GLenum error = glGetError();
    if (error != GL_NO_ERROR) {
        printf("Failed to clear framebuffer: %x\n", error);
        abort();
    }
}

void GLRenderer::useGeometry(const std::shared_ptr<Geometry>& geometry)
{
    auto glGeometry = std::static_pointer_cast<GLGeometry, Geometry>(geometry);

    glBindVertexArray(glGeometry->vao());
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, glGeometry->indexBuffer());

    useInstanceBuffer();

    m_currentGeometry = geometry;
}

void GLRenderer::setupCamera(const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix)
{
    m_shaderUniforms.viewMatrix = viewMatrix;
    m_shaderUniforms.projectionMatrix = projectionMatrix;

    // Update uniforms.
    GLuint uniformBuffer = m_perFrameBuffers[m_currentFrameBuffer].uniformBuffer;
    ShaderUniforms* mapped = reinterpret_cast<ShaderUniforms*>(glMapNamedBufferRange(uniformBuffer, 0, sizeof m_shaderUniforms, GL_MAP_WRITE_BIT));

    GLenum error = glGetError();
    if (error != GL_NO_ERROR) {
        printf("Failed to map uniforms: %x\n", error);
        abort();
    }

    memcpy(mapped, &m_shaderUniforms, sizeof m_shaderUniforms);
    glUnmapNamedBuffer(uniformBuffer);

    glBindBufferBase(GL_UNIFORM_BUFFER, 0, uniformBuffer);
    glUniformBlockBinding(m_program, 0, 0);

    // Map instance buffer
    m_instanceBuffer.data = reinterpret_cast<glm::mat4*>(
        glMapNamedBufferRange(m_perFrameBuffers[m_currentFrameBuffer].instanceBuffer, 0, sizeof(glm::mat4) * m_instanceBuffer.maxInstances, GL_MAP_WRITE_BIT));

    error = glGetError();
    if (error != GL_NO_ERROR) {
        printf("Failed to map instance buffer: %x\n", error);
        abort();
    }
}

void GLRenderer::setViewport(int32_t x, int32_t y, int32_t width, int32_t height)
{
    glScissor(x, y, width, height);
    glViewport(x, y, width, height);

    GLenum error = glGetError();
    if (error != GL_NO_ERROR) {
        printf("Failed to set viewport: %x\n", error);
        abort();
    }
}

void GLRenderer::drawGrid()
{
    glDisable(GL_DEPTH_TEST);
    glDepthMask(GL_FALSE);

    glUseProgram(m_gridProgram);
    glUniformBlockBinding(m_gridProgram, 0, 0);
    glDrawElements(GL_TRIANGLES, m_currentGeometry->indexCount(), GL_UNSIGNED_INT, nullptr);

    glDepthMask(GL_TRUE);
    glEnable(GL_DEPTH_TEST);
}

void GLRenderer::queueObjects(const std::vector<glm::mat4>& matrices)
{
    if (m_instanceBuffer.queued > 0) {
        dispatchDraws();
    }

    memcpy(m_instanceBuffer.data, matrices.data(), sizeof(glm::mat4) * matrices.size());
    m_instanceBuffer.queued = static_cast<int32_t>(matrices.size());
}

void GLRenderer::dispatchDraws()
{
    if (m_instanceBuffer.queued == 0) {
        return;
    }

    glUnmapNamedBuffer(m_perFrameBuffers[m_currentFrameBuffer].instanceBuffer);

    GLenum error = glGetError();
    if (error != GL_NO_ERROR) {
        printf("Failed to unmap instance buffer: %x\n", error);
        abort();
    }

    glUseProgram(m_program);
    glDrawElementsInstanced(GL_TRIANGLES, m_currentGeometry->indexCount(), GL_UNSIGNED_INT, nullptr, m_instanceBuffer.queued);
    m_instanceBuffer.queued = 0;
}

void GLRenderer::advance() { m_currentFrameBuffer = (m_currentFrameBuffer + 1) % m_perFrameBuffers.size(); }

void GLRenderer::compileShaders()
{
    compileShader();
    compileGridShader();
}

void GLRenderer::compileShader()
{
    const char* vertexSource =
        "#version 430\n"
        "layout(location = 0) in vec3 position;\n"
        "layout(location = 1) in vec3 normal;\n"
        "layout(location = 2) in vec4 worldMatrix0;\n"
        "layout(location = 3) in vec4 worldMatrix1;\n"
        "layout(location = 4) in vec4 worldMatrix2;\n"
        "layout(location = 5) in vec4 worldMatrix3;\n"
        "\n"
        "layout(std140, binding = 0) uniform Matrices {\n"
        "    mat4 viewMatrix;\n"
        "    mat4 projectionMatrix;\n"
        "};\n"
        "\n"
        "out vec3 vNormal;"
        "\n"
        "void main() {\n"
        "    mat4 worldMat = mat4(worldMatrix0, worldMatrix1, worldMatrix2, worldMatrix3);\n"
        "    vNormal = (worldMat * vec4(normal, 0)).xyz;\n"
        "    gl_Position = projectionMatrix * viewMatrix * worldMat * vec4(position, 1);\n"
        "}\n";

    const char* fragmentSource =
        "#version 430\n"
        "\n"
        "in vec3 vNormal;\n"
        "out vec4 oColor;\n"
        "void main() {\n"
        "    oColor = vec4(pow(vNormal, vec3(1 / 2.2)), 1);\n"
        "}\n";

    GLuint vertexShader = ::compileShader(GL_VERTEX_SHADER, vertexSource, "vertex shader");
    GLuint fragmentShader = ::compileShader(GL_FRAGMENT_SHADER, fragmentSource, "fragment shader");

    m_program = linkProgram(vertexShader, fragmentShader, "program");
}

void GLRenderer::compileGridShader()
{
    const char* vertexSource =
        "#version 430\n"
        "layout(location = 0) in vec3 position;\n"
        "layout(location = 1) in vec3 normal;\n"
        "\n"
        "layout(std140, binding = 0) uniform Matrices {\n"
        "    mat4 viewMatrix;\n"
        "    mat4 projectionMatrix;\n"
        "};\n"
        "\n"
        "out vec3 vPosition;\n"
        "\n"
        "void main() {\n"
        "    vPosition = position + 0.5;\n"
        "\n"
        "    mat4 view = viewMatrix;\n"
        "    view[3][0] = 0;\n"
        "    view[3][1] = 0;\n"
        "    view[3][2] = 0;\n"
        "    gl_Position = projectionMatrix * view * vec4(position, 1);\n"
        "}\n";

    const char* fragmentSource =
        "#version 430\n"
        "\n"
        "in vec3 vPosition;\n"
        "\n"
        "out vec4 oColor;\n"
        "\n"
        "float grid(float coordinate, float size) {\n"
        "    float inRange = step(0.001, 1.0 - coordinate) * step(0.001, coordinate);\n"
        "    float x = (coordinate * size);\n"
        "    return inRange * step(0.25, x - floor(x)) * step(0.25, 1.0 - (x - floor(x)));\n"
        "}\n"
        "\n"
        "void main() {\n"
        "    float x = grid(vPosition.x, 4.0f);\n"
        "    float y = grid(vPosition.y, 4.0f);\n"
        "    float z = grid(vPosition.z, 4.0f);\n"
        "    float value = clamp(x + y + z, 0.0, 1.0);\n"
        "\n"
        "    vec3 color = vec3(step(0.999, vPosition.z) * mix(0.65, 0.427, value));\n"
        "    color += step(0.999, 1.0 - vPosition.z) * mix(1.0, 0.0, value);\n"
        "\n"
        "    color += step(0.999, vPosition.x) * mix(vec3(1, 0, 0), vec3(0, 1, 0), value);\n"
        "    color += step(0.999, 1.0 - vPosition.x) * mix(vec3(1, 0, 1), vec3(0, 0, 1), value);\n"
        "\n"
        "    color += step(0.999, vPosition.y) * mix(vec3(1, 1, 0), vec3(0, 1, 1), value);\n"
        "    color += step(0.999, 1.0 - vPosition.y) * mix(vec3(0.25, 0, 0.392), vec3(0, 0.392, 0.129), value);\n"
        "\n"
        "    oColor = vec4(pow(color, vec3(1 / 2.2)), 1);\n"
        "}\n";

    GLuint vertexShader = ::compileShader(GL_VERTEX_SHADER, vertexSource, "grid vertex shader");
    GLuint fragmentShader = ::compileShader(GL_FRAGMENT_SHADER, fragmentSource, "grid fragment shader");

    m_gridProgram = linkProgram(vertexShader, fragmentShader, "grid program");
    glUseProgram(m_gridProgram);
}

void GLRenderer::createPerFrameBuffers()
{
    for (int i = 0; i < 4 * 4; ++i) {
        GLuint instanceBuffer = createInstanceBuffer();
        GLuint uniformBuffer = createUniformBuffer();

        m_perFrameBuffers.push_back({
            uniformBuffer,
            instanceBuffer,
        });
    }
}

GLuint GLRenderer::createInstanceBuffer()
{
    const int32_t maxInstances = 5000;
    m_instanceBuffer.maxInstances = maxInstances;

    GLuint buffer = 0;
    glGenBuffers(1, &buffer);
    glBindBuffer(GL_ARRAY_BUFFER, buffer);
    glBufferStorage(GL_ARRAY_BUFFER, maxInstances * sizeof(glm::mat4), nullptr, GL_MAP_WRITE_BIT);

    GLenum error = glGetError();
    if (error != GL_NO_ERROR) {
        printf("Failed to create instance buffer: %x", error);
        abort();
    }

    return buffer;
}

GLuint GLRenderer::createUniformBuffer()
{
    GLuint buffer = 0;
    glGenBuffers(1, &buffer);
    glBindBuffer(GL_UNIFORM_BUFFER, buffer);
    glBufferStorage(GL_UNIFORM_BUFFER, sizeof m_shaderUniforms, nullptr, GL_MAP_WRITE_BIT);

    GLenum error = glGetError();
    if (error != GL_NO_ERROR) {
        printf("Failed to create uniform buffer: %x", error);
        abort();
    }
    return buffer;
}

void GLRenderer::useInstanceBuffer()
{
    // Set up vertex attributes for instancing.
    glBindBuffer(GL_ARRAY_BUFFER, m_perFrameBuffers[m_currentFrameBuffer].instanceBuffer);

    glEnableVertexAttribArray(2);
    glVertexAttribPointer(2, 4, GL_FLOAT, false, sizeof(glm::mat4), 0);
    glVertexAttribDivisor(2, 1);

    glEnableVertexAttribArray(3);
    glVertexAttribPointer(3, 4, GL_FLOAT, false, sizeof(glm::mat4), reinterpret_cast<void*>(static_cast<uintptr_t>(sizeof(float) * 4)));
    glVertexAttribDivisor(3, 1);

    glEnableVertexAttribArray(4);
    glVertexAttribPointer(4, 4, GL_FLOAT, false, sizeof(glm::mat4), reinterpret_cast<void*>(static_cast<uintptr_t>(sizeof(float) * 8)));
    glVertexAttribDivisor(4, 1);

    glEnableVertexAttribArray(5);
    glVertexAttribPointer(5, 4, GL_FLOAT, false, sizeof(glm::mat4), reinterpret_cast<void*>(static_cast<uintptr_t>(sizeof(float) * 12)));
    glVertexAttribDivisor(5, 1);
}

void GLRenderer::createDummyWindow()
{
    const int width = 800;
    const int height = 600;

    const wchar_t* windowClassName = L"Varjo";
    const wchar_t* windowTitle = L"Varjo";

    HINSTANCE hInstance = GetModuleHandle(nullptr);

    WNDCLASS windowClass{};
    windowClass.style = CS_HREDRAW | CS_VREDRAW;
    windowClass.lpfnWndProc = wndProc;
    windowClass.cbClsExtra = 0;
    windowClass.cbWndExtra = 0;
    windowClass.hInstance = hInstance;
    windowClass.hIcon = LoadIcon(hInstance, NULL);
    windowClass.hCursor = LoadCursor(NULL, IDC_ARROW);
    windowClass.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
    windowClass.lpszMenuName = NULL;
    windowClass.lpszClassName = windowClassName;

    if (!RegisterClassW(&windowClass)) {
        printf("RegisterClassEx failed: %s\n", getLastErrorString());
        abort();
    }

    m_hwnd = CreateWindowW(windowClassName, windowTitle, WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, CW_USEDEFAULT, width, height, NULL, NULL, hInstance, NULL);
    if (!m_hwnd) {
        printf("CreateWindow failed: %s\n", getLastErrorString());
        abort();
    }

    UpdateWindow(m_hwnd);
}

namespace
{
GLuint compileShader(GLenum type, const char* source, const char* name)
{
    GLuint shader = glCreateShader(type);
    glShaderSource(shader, 1, &source, nullptr);
    glCompileShader(shader);

    GLint status = 0;
    glGetShaderiv(shader, GL_COMPILE_STATUS, &status);
    if (status != GL_TRUE) {
        GLint infoSize = 0;
        glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &infoSize);

        std::vector<char> infoLog(infoSize);
        glGetShaderInfoLog(shader, infoSize, &infoSize, infoLog.data());

        printf("Failed to compile '%s': \n", name);
        printf("%s\n", infoLog.data());
        abort();
    }

    return shader;
}

GLuint linkProgram(GLuint vertexShader, GLuint fragmentShader, const char* name)
{
    GLuint program = glCreateProgram();
    glAttachShader(program, vertexShader);
    glAttachShader(program, fragmentShader);
    glLinkProgram(program);

    GLint status = 0;
    glGetProgramiv(program, GL_LINK_STATUS, &status);
    if (status == GL_FALSE) {
        GLint infoLength = 0;
        glGetProgramiv(program, GL_INFO_LOG_LENGTH, &infoLength);

        std::vector<char> info(infoLength);
        glGetProgramInfoLog(program, infoLength, nullptr, info.data());

        printf("Failed to link %s: %s", name, info.data());
        abort();
    }

    return program;
}

LRESULT CALLBACK wndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) {
        case WM_DESTROY: PostQuitMessage(0); break;
        default: return DefWindowProc(hwnd, message, wParam, lParam); break;
    }

    return 0;
}
}  // namespace
