#include <stdio.h>

#include <glm/gtc/type_ptr.hpp>
#include <glm/gtc/matrix_transform.hpp>

#include "GazeTracking.hpp"

GazeTracking::GazeTracking(varjo_Session* session)
    : m_session(session)
    , m_initialized(false)
    , m_position(0, 0, 0)
{
}

void GazeTracking::init()
{
    if (m_initialized) return;

    // Check that gaze tracking is allowed
    if (!varjo_IsGazeAllowed(m_session)) {
        printf("Gaze tracking is not allowed!\n");
        return;
    }

    // Initialize gaze tracking
    varjo_GazeInit(m_session);
    varjo_Error err = varjo_GetError(m_session);
    if (err != varjo_NoError) {
        printf("Failed to initialize Gaze: %s", varjo_GetErrorDesc(err));
        return;
    }

    m_initialized = true;
}

void GazeTracking::requestCalibration()
{
    if (!m_initialized) return;

    // Request gaze calibration
    printf("Gaze calibration requested\n");
    varjo_RequestGazeCalibration(m_session);
}

bool GazeTracking::update()
{
    if (!m_initialized) return false;

    // Get gaze and check that it is valid
    varjo_Gaze gaze = varjo_GetGaze(m_session);
    if (gaze.status == varjo_GazeStatus_Invalid) return false;

    // Calculate relative gaze vector
    glm::vec3 dir = glm::make_vec3(gaze.gaze.forward);
    dir *= gaze.focusDistance;  // Position it on distance where user looks at

    // In gaze coordinates positive Z is in front of user, flip Z axis for this environment
    dir.z *= -1.0f;


    // Position gaze in relative to user pose
    varjo_Matrix m = varjo_FrameGetPose(m_session, varjo_PoseType_Center);
    glm::mat4 mat{glm::make_mat4(m.value)};

    // Calculate gaze position in world coordinates
    m_position = glm::vec3(mat * glm::vec4(dir, 1));
    return true;
}
