#include <stdlib.h>
#include <stdio.h>

#include <glm/gtc/type_ptr.hpp>
#include <glm/gtc/matrix_transform.hpp>

#include "IRenderer.hpp"
#include "GeometryGenerator.hpp"

namespace
{
glm::mat4 doubleMatrixToGLMMatrix(const double* dMatrix);
const char* getErrorMsg(const unsigned long error);
void calculateWorldMatrices(std::vector<glm::mat4>& worldMatrices, const std::vector<IRenderer::Object>& objects);
}  // namespace

IRenderer::IRenderer(varjo_Session* session, bool useMultiTexture)
    : m_session(session)
    , m_useMultiTexture(useMultiTexture)
{
}

IRenderer::~IRenderer() {}

bool IRenderer::init()
{
    // Initialize rendering API specific Varjo graphics API.
    if (!initVarjo()) {
        return false;
    }

    if (m_useMultiTexture) {
        std::vector<varjo_Viewport> viewports(m_graphicsInfo->viewCount);
        varjo_LayoutDefaultViewports(m_session, viewports.data());

        for (int i = 0; i < m_graphicsInfo->viewCount; ++i) {
            const auto viewport = viewports[i];
            m_renderTargets.push_back(createRenderTarget(viewport.width, viewport.height, varjo_TextureFormat_R8G8B8A8_SRGB));
        }

    } else {
        // Create a render target for each swap chain texture.
        for (int i = 0; i < m_graphicsInfo->swapChainTextureCount; ++i) {
            m_renderTargets.push_back(
                createRenderTarget(m_graphicsInfo->swapChainTextureWidth, m_graphicsInfo->swapChainTextureHeight, m_graphicsInfo->swapChainTextures[i]));
        }
    }

    // Cube geometry for the background grid.
    m_cubeGeometry = GeometryGenerator::generateCube(shared_from_this(), 1.0f, 1.0f, 1.0f);

    return true;
}

void IRenderer::updateViewportLayout(varjo_SubmitInfo* submitInfo)
{
    if (!m_useMultiTexture) {
        return;
    }

    for (int i = 0; i < m_graphicsInfo->viewCount; ++i) {
        submitInfo->viewports[i].x = 0;
        submitInfo->viewports[i].y = 0;
    }
}

void IRenderer::drawObjects(const std::vector<Object>& objects, const std::vector<glm::mat4>& matrices)
{
    if (objects.size()) {
        // Assume all of the objects use the same geometry (for now).
        useGeometry(objects[0].geometry);
        // Queue objects and render
        queueObjects(matrices);
        dispatchDraws();
    }
}

void IRenderer::render(varjo_FrameInfo* frameInfo, varjo_SubmitInfo* submitInfo, int32_t swapChainIndex, const std::vector<Object>& objects,
    const std::vector<Object>& controllerObjects)
{
    // Don't render if there are no objects.
    if (objects.size() == 0 && controllerObjects.size() == 0) {
        return;
    }

    // Setup render target once if using the swap chain.
    if (!m_useMultiTexture) {
        // Use the current swap chain texture.
        m_currentRenderTarget = m_renderTargets[swapChainIndex];
        useRenderTarget(m_currentRenderTarget);

        // Clear the whole texture to black.
        setViewport(0, 0, m_currentRenderTarget->width(), m_currentRenderTarget->height());
        clearRenderTarget(m_currentRenderTarget, 0, 0, 0, 1);
    }

    // Calculate object world matrices.
    calculateWorldMatrices(m_objectWorldMatrices, objects);
    calculateWorldMatrices(m_controllerObjectWorldMatrices, controllerObjects);

    // Render all active views.
    for (int i = 0; i < m_graphicsInfo->viewCount; ++i) {
        varjo_ViewInfo& view = frameInfo->views[i];
        if (!view.enabled) {
            continue;  // Skip a view if it is not enabled.
        }

        // Setup each render target when using multiple textures.
        if (m_useMultiTexture) {
            m_currentRenderTarget = m_renderTargets[i];
            useRenderTarget(m_currentRenderTarget);

            setViewport(0, 0, m_currentRenderTarget->width(), m_currentRenderTarget->height());
            clearRenderTarget(m_currentRenderTarget, 1, 0, 0, 1);
        } else {
            // Set up the viewport.
            varjo_Viewport viewport = submitInfo->viewports[i];
            setViewport(viewport.x, viewport.y, viewport.width, viewport.height);
        }

        // Change the near and far clip distances
        varjo_UpdateClipPlaneDistances(m_session, view.projectionMatrix, 0.1, 1000.0);

        // Setup the view and projection matrices.
        glm::mat4 viewMatrix = doubleMatrixToGLMMatrix(view.viewMatrix);
        glm::mat4 projectionMatrix = doubleMatrixToGLMMatrix(view.projectionMatrix);

        setupCamera(viewMatrix, projectionMatrix);

        // Draw the background grid.
        useGeometry(m_cubeGeometry);
        drawGrid();

        drawObjects(objects, m_objectWorldMatrices);
        drawObjects(controllerObjects, m_controllerObjectWorldMatrices);

        advance();
        // Submit the right render target.
        submitInfo->textures[i] = m_currentRenderTarget->toVarjoTexture();
    }
}

const char* IRenderer::getLastErrorString()
{
    DWORD err = GetLastError();
    return err ? getErrorMsg(err) : "";
}

namespace
{
void calculateWorldMatrices(std::vector<glm::mat4>& worldMatrices, const std::vector<IRenderer::Object>& objects)
{
    worldMatrices.resize(objects.size());

    size_t numObjects = objects.size();
    for (size_t i = 0; i < numObjects; ++i) {
        auto& object = objects[i];

        glm::mat4 matrix = glm::toMat4(object.orientation);
        matrix[3][0] = object.position.x;
        matrix[3][1] = object.position.y;
        matrix[3][2] = object.position.z;
        matrix = glm::scale(matrix, object.scale);

        worldMatrices[i] = matrix;
    }
}

glm::mat4 doubleMatrixToGLMMatrix(const double* dMatrix)
{
    glm::mat4 result{};
    float* dst = glm::value_ptr(result);

    for (int i = 0; i < 16; ++i) dst[i] = static_cast<float>(dMatrix[i]);

    return result;
}

const char* getErrorMsg(const unsigned long error)
{
    static char buffer[MAX_PATH];

    // Call FormatMessage to get error string from winapi
    LPWSTR _buffer = nullptr;
    DWORD size = FormatMessageW(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, nullptr, error,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPWSTR)&_buffer, 0, nullptr);

    if (size > 0) {
        std::vector<char> descBuffer(size * 2);
        size_t convertCount;
        wcstombs_s(&convertCount, descBuffer.data(), size * 2, _buffer, size);
        sprintf_s(buffer, "0x%X: %s", error, descBuffer.data());
    } else {
        sprintf_s(buffer, "0x%X", error);
    }
    LocalFree(_buffer);

    // trim leading newlines, some error messages might contain them
    int i = (int)(strlen(buffer) - 1);
    for (; i >= 0 && (buffer[i] == '\n' || buffer[i] == '\r'); i--) {
        buffer[i] = '\0';
    }

    return buffer;
}
}  // namespace
