#pragma once

#include <vector>
#include <memory>
#include <Varjo.h>

#define GLM_ENABLE_EXPERIMENTAL
#include <glm/gtx/quaternion.hpp>

#include "Geometry.hpp"
#include "RenderTarget.hpp"

class IRenderer : public std::enable_shared_from_this<IRenderer>
{
public:
    struct Object {
        std::shared_ptr<Geometry> geometry;
        glm::vec3 position;
        glm::vec3 scale;
        glm::quat orientation;
    };

    IRenderer(varjo_Session* session, bool useMultiTexture);
    virtual ~IRenderer();

    bool init();

    void updateViewportLayout(varjo_SubmitInfo* submitInfo);

    virtual std::shared_ptr<Geometry> createGeometry(uint32_t vertexCount, uint32_t indexCount) = 0;
    virtual std::shared_ptr<RenderTarget> createRenderTarget(int32_t width, int32_t height, varjo_TextureFormat format) = 0;
    virtual std::shared_ptr<RenderTarget> createRenderTarget(int32_t width, int32_t height, varjo_Texture texture) = 0;

    void render(varjo_FrameInfo* frameInfo, varjo_SubmitInfo* submitInfo, int32_t swapChainIndex, const std::vector<Object>& objects,
        const std::vector<Object>& controllerObjects);

protected:
    // Initialize the Varjo graphics API.
    virtual bool initVarjo() = 0;

    virtual void useRenderTarget(const std::shared_ptr<RenderTarget>& renderTarget) = 0;
    virtual void clearRenderTarget(const std::shared_ptr<RenderTarget>& renderTarget, float r, float g, float b, float a) = 0;

    virtual void useGeometry(const std::shared_ptr<Geometry>& geometry) = 0;

    virtual void setupCamera(const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix) = 0;
    virtual void setViewport(int32_t x, int32_t y, int32_t width, int32_t height) = 0;

    // Draw the background grid.
    virtual void drawGrid() = 0;

    // Queue objects to be drawn. Uses the current geometry.
    virtual void queueObjects(const std::vector<glm::mat4>& matrices) = 0;

    // Dispatch the drawing of the queued objects.
    virtual void dispatchDraws() = 0;

    virtual void advance() = 0;

    void drawObjects(const std::vector<Object>& objects, const std::vector<glm::mat4>& matrices);

    const char* getLastErrorString();

    varjo_Session* m_session = nullptr;
    varjo_GraphicsInfo* m_graphicsInfo = nullptr;

    // Temp vector for object world matrices.
    std::vector<glm::mat4> m_objectWorldMatrices;

    // Temp vector for object world matrices for controllers.
    std::vector<glm::mat4> m_controllerObjectWorldMatrices;

    // Currently active geometry.
    std::shared_ptr<Geometry> m_currentGeometry;

    // Currently active render target.
    std::shared_ptr<RenderTarget> m_currentRenderTarget;

    std::shared_ptr<Geometry> m_cubeGeometry;
    std::vector<std::shared_ptr<RenderTarget>> m_renderTargets;

    bool m_useMultiTexture;
};
