#include <stdio.h>
#include "RenderTarget.hpp"
#include "D3D11Renderer.hpp"

RenderTarget::RenderTarget(int32_t width, int32_t height)
    : m_width(width)
    , m_height(height)
{
}

RenderTarget::~RenderTarget() {}

D3D11RenderTarget::D3D11RenderTarget(D3D11Renderer* renderer, int32_t width, int32_t height, ID3D11Texture2D* backBuffer)
    : RenderTarget(width, height)
    , m_backBuffer(nullptr)
{
    create(renderer, backBuffer);
}

D3D11RenderTarget::D3D11RenderTarget(D3D11Renderer* renderer, int32_t width, int32_t height, varjo_TextureFormat format)
    : RenderTarget(width, height)
    , m_backBuffer(nullptr)
{
    ID3D11Texture2D* backBuffer = nullptr;

    DXGI_FORMAT dxgiFormat{};
    switch (format) {
        case varjo_TextureFormat_R8G8B8A8_SRGB: dxgiFormat = DXGI_FORMAT_R8G8B8A8_UNORM_SRGB; break;
        case varjo_TextureFormat_B8G8R8A8_SRGB: dxgiFormat = DXGI_FORMAT_B8G8R8A8_UNORM_SRGB; break;
    }

    D3D11_TEXTURE2D_DESC textureDesc{};
    textureDesc.Width = width;
    textureDesc.Height = height;
    textureDesc.MipLevels = 1;
    textureDesc.ArraySize = 1;
    textureDesc.SampleDesc.Count = 1;
    textureDesc.SampleDesc.Quality = 0;
    textureDesc.Format = dxgiFormat;
    textureDesc.Usage = D3D11_USAGE_DEFAULT;
    textureDesc.BindFlags = D3D11_BIND_RENDER_TARGET;

    HRESULT result = renderer->dxDevice()->CreateTexture2D(&textureDesc, nullptr, &backBuffer);
    if (result != S_OK) {
        printf("Failed to create DX texture: %d\n", result);
        abort();
    }

    create(renderer, backBuffer);
    backBuffer->Release();
}

D3D11RenderTarget::~D3D11RenderTarget()
{
    m_depthStencilView->Release();
    m_renderTargetView->Release();
    m_backBuffer->Release();
}

varjo_Texture D3D11RenderTarget::toVarjoTexture()
{
    // Converts a ID3D11Texture2D* to varjo_Texture.
    return varjo_FromD3D11Texture(m_backBuffer);
}

void D3D11RenderTarget::create(D3D11Renderer* renderer, ID3D11Texture2D* backBuffer)
{
    m_backBuffer = backBuffer;

    CD3D11_TEXTURE2D_DESC depthStencilDesc(
        DXGI_FORMAT_D24_UNORM_S8_UINT, static_cast<UINT>(m_width), static_cast<UINT>(m_height), 1, 1, D3D11_BIND_DEPTH_STENCIL);

    ID3D11Texture2D* depthStencil = nullptr;
    HRESULT result = renderer->dxDevice()->CreateTexture2D(&depthStencilDesc, nullptr, &depthStencil);
    if (result != S_OK) {
        printf("Failed to create DX texture: %d", result);
        abort();
    }

    CD3D11_DEPTH_STENCIL_VIEW_DESC depthStencilViewDesc(D3D11_DSV_DIMENSION_TEXTURE2D);
    result = renderer->dxDevice()->CreateDepthStencilView(depthStencil, &depthStencilViewDesc, &m_depthStencilView);
    if (result != S_OK) {
        printf("Failed to create depth stencil view: %d", result);
        abort();
    }

    depthStencil->Release();

    D3D11_RENDER_TARGET_VIEW_DESC rtvDesc{};
    rtvDesc.Format = DXGI_FORMAT_R8G8B8A8_UNORM_SRGB;
    rtvDesc.ViewDimension = D3D11_RTV_DIMENSION_TEXTURE2D;
    rtvDesc.Texture2D.MipSlice = 0;

    result = renderer->dxDevice()->CreateRenderTargetView(m_backBuffer, &rtvDesc, &m_renderTargetView);
    if (result != S_OK) {
        printf("Failed to create render target view: %d", result);
        abort();
    }

    m_backBuffer->AddRef();
}

GLRenderTarget::GLRenderTarget(int32_t width, int32_t height, GLuint texture)
    : RenderTarget(width, height)
    , m_frameBuffer{0}
    , m_colorBuffer{0}
    , m_depthBuffer{0}
{
    create(texture);
}

GLRenderTarget::GLRenderTarget(int32_t width, int32_t height, varjo_TextureFormat format)
    : RenderTarget(width, height)
{
    GLenum glFormat{};
    switch (format) {
        case varjo_TextureFormat_R8G8B8A8_SRGB: glFormat = GL_RGBA8; break;
        case varjo_TextureFormat_B8G8R8A8_SRGB: abort(); break;
    }

    GLuint colorBuffer{};
    glGenTextures(1, &colorBuffer);
    glBindTexture(GL_TEXTURE_2D, colorBuffer);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, nullptr);

    GLenum error = glGetError();
    if (error != GL_NO_ERROR) {
        printf("Failed to create GL color buffer: %x", error);
        abort();
    }

    glBindTexture(GL_TEXTURE_2D, 0);
    create(colorBuffer);
}

GLRenderTarget::~GLRenderTarget()
{
    glDeleteRenderbuffers(1, &m_depthBuffer);
    glDeleteFramebuffers(1, &m_frameBuffer);
}

varjo_Texture GLRenderTarget::toVarjoTexture()
{
    // Converts an OpenGL texture handle to varjo_Texture.
    return varjo_FromGLTexture(m_colorBuffer);
}

void GLRenderTarget::create(GLuint colorBuffer)
{
    m_colorBuffer = colorBuffer;

    glGenFramebuffers(1, &m_frameBuffer);
    glBindFramebuffer(GL_FRAMEBUFFER, m_frameBuffer);

    glGenRenderbuffers(1, &m_depthBuffer);
    glBindRenderbuffer(GL_RENDERBUFFER, m_depthBuffer);
    glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT, m_width, m_height);

    GLenum error = glGetError();
    if (error != GL_NO_ERROR) {
        printf("Failed to create render buffer: %x", error);
        abort();
    }

    glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, m_depthBuffer);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, m_colorBuffer, 0);

    GLenum drawBuffers[1] = {GL_COLOR_ATTACHMENT0};
    glDrawBuffers(1, drawBuffers);

    GLenum status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
    if (status != GL_FRAMEBUFFER_COMPLETE) {
        printf("Incomplete framebuffer: %x\n", status);
        abort();
    }
}
