/**
 * Varjo SDK example.
 *
 * main.cpp:
 *      Contains the main rendering loop and all of the frame logic that
 *      is not related to graphics.
 *
 *      This application renders a number of donuts and a background grid
 *      to stress test the Varjo API.
 *
 * IRenderer.hpp:
 *      Contains an abstract renderer class that takes care of actual rendering.
 *      Implementation can be found from D3D11Renderer.hpp and GLRenderer.hpp.
 *
 * Copyright (C) 2018 Varjo Technologies Ltd.
 */

#define GLM_ENABLE_EXPERIMENTAL
#include <glm/vec3.hpp>
#include <glm/gtx/quaternion.hpp>

#include <Varjo.h>
#include <Varjo_events.h>

#include "OpenVRTracker.hpp"
#include "Profiler.hpp"
#include "Arguments.hpp"
#include "GLRenderer.hpp"
#include "D3D11Renderer.hpp"
#include "GeometryGenerator.hpp"
#include "GazeTracking.hpp"

void printHelp();
bool gotKey();

struct ObjectRotation {
    glm::vec3 axis;
    float speed;
};

void createObjects(std::shared_ptr<IRenderer> renderer, std::vector<IRenderer::Object>& object, std::vector<ObjectRotation>& rotations);

void createController(std::shared_ptr<IRenderer> renderer, IRenderer::Object& controllerObject);
void createGaze(std::shared_ptr<IRenderer> renderer, IRenderer::Object& gazeObject);

int main(int argc, char** argv)
{
    Arguments arguments(argc, argv);
    if (arguments.hasArgument("help")) {
        printHelp();
        exit(0);
    }

    bool useMultiTexture = arguments.hasArgument("multi-texture");
    bool disableAnimation = arguments.hasArgument("disable-animation");
    bool enableProfiling = arguments.hasArgument("profile-start-frame") && arguments.hasArgument("profile-frame-count");
    bool useGaze = arguments.hasArgument("gaze");

    printf("Multi texture %s\n", useMultiTexture ? "enabled" : "disabled");
    printf("Animation %s\n", disableAnimation ? "disabled" : "enabled");
    printf("Gaze %s\n", useGaze ? "in use" : "not used");

    int32_t profileStartFrame = arguments.value<int32_t>("profile-start-frame");
    int32_t profileFrameCount = arguments.value<int32_t>("profile-frame-count");

    if (enableProfiling) {
        printf("Profile:\n");
        printf("  start frame: %d\n", profileStartFrame);
        printf("  frame count: %d\n", profileFrameCount);
    }

    Profiler profiler;

    if (!varjo_IsAvailable()) {
        printf("Varjo system not available. Exiting...");
        exit(1);
    }

    // Initialize the varjo session
    varjo_Session* session = varjo_SessionInit();

    // Check if there was any errors while initializing.
    varjo_Error error = varjo_GetError(session);
    if (error != varjo_NoError) {
        printf("Failed to initialize Varjo session: %s", varjo_GetErrorDesc(error));
        exit(1);
    }

    // Choose the rendering API.
    std::string rendererName = arguments.hasArgument("renderer") ? arguments.value<std::string>("renderer") : "d3d11";

    std::shared_ptr<IRenderer> renderer;
    if (rendererName == "gl") {
        renderer = std::make_shared<GLRenderer>(session, useMultiTexture);
    } else if (rendererName == "d3d11") {
        renderer = std::make_shared<D3D11Renderer>(session, useMultiTexture);
    } else {
        printf("Unknown renderer '%s'", rendererName.c_str());
        exit(1);
    }

    // Initialize.
    // Calls varjo_*Init and fetches all swap chain textures.
    renderer->init();

    OpenVRTracker openVRTracker;
    openVRTracker.init();

    // Initialize gaze tracking when needed
    GazeTracking gaze(session);
    if (useGaze) gaze.init();

    std::vector<IRenderer::Object> donutObjects;
    std::vector<ObjectRotation> rotations;
    IRenderer::Object controllerObject;
    IRenderer::Object gazeObject;

    createObjects(renderer, donutObjects, rotations);
    createController(renderer, controllerObject);
    createGaze(renderer, gazeObject);

    // Frame info is used for per-frame view and projection matrices.
    varjo_FrameInfo* frameInfo = varjo_CreateFrameInfo(session);

    // Submit info contains all of the information required to submit a frame.
    // Contains default viewport configuration.
    varjo_SubmitInfo* submitInfo = varjo_CreateSubmitInfo(session);
    submitInfo->flags = useMultiTexture ? 0 : varjo_SubmitFlag_Async;

    // Update the viewport layout. Swap chain uses the default layout but multitexture needs a custom layout.
    renderer->updateViewportLayout(submitInfo);

    // A struct to hold Varjo event data.
    varjo_Event evt{};
    varjo_Nanoseconds lastFrameTime = varjo_GetCurrentTime(session);

    int32_t frameNumber = 0;

    bool visible = true;
    while (!gotKey()) {
        // Poll Varjo events.
        while (varjo_PollEvent(session, &evt)) {
            switch (evt.header.type) {
                case varjo_EventType_Visibility: {
                    // Don't render anything when we are hidden.
                    visible = evt.data.visibility.visible;
                    printf("Visible %s\n", visible ? "true" : "false");
                } break;
                case varjo_EventType_Foreground: {
                    printf("In foreground %s\n", evt.data.foreground.isForeground ? "true" : "false");
                } break;
                case varjo_EventType_HeadsetStandbyStatus: {
                    printf("Headset on standby %s\n", evt.data.headsetStandbyStatus.onStandby ? "true" : "false");
                } break;
                case varjo_EventType_Button: {
                    if (evt.data.button.buttonId == varjo_ButtonId_Application && evt.data.button.pressed) {
                        // Request gaze calibration when button is pressed
                        gaze.requestCalibration();
                    }
                } break;
            }
        }

        if (visible) {
            // Wait for a perfect time to render the frame.
            varjo_WaitSync(session, frameInfo);
            // Update the tracking position for hand controllers.
            openVRTracker.update();

            if (enableProfiling && frameNumber >= profileStartFrame) {
                profiler.addSample();

                if (profiler.sampleCount() == 0) {
                    printf("Start profiling.\n");
                }
            }

            float time = (frameInfo->displayTime - lastFrameTime) / 1000000000.0f;

            if (!disableAnimation) {
                size_t numObjects = donutObjects.size();
                for (size_t i = 0; i < numObjects; ++i) {
                    auto& object = donutObjects[i];
                    auto& rotate = rotations[i];

                    glm::quat rot = glm::angleAxis(rotate.speed * time, rotate.axis);
                    object.orientation = glm::normalize(object.orientation * rot);
                }
            }

            std::vector<IRenderer::Object> controllerObjects;

            for (int i = 0; i < openVRTracker.getControllerCount(); ++i) {
                IRenderer::Object newObject = controllerObject;
                newObject.position = openVRTracker.getControllerPosition(i);
                newObject.orientation = openVRTracker.getControllerOrientation(i);
                if (openVRTracker.isButtonTouched(i)) newObject.scale = newObject.scale * glm::vec3{1.5, 1.5, 1.5};
                if (openVRTracker.isButtonPressed(i)) newObject.scale = newObject.scale * glm::vec3{1.5, 1.5, 1.5};
                controllerObjects.push_back(newObject);
            }

            // Add object where user is looking when gaze data is valid
            if (gaze.update()) {
                IRenderer::Object newObject = gazeObject;
                newObject.position = gaze.getPosition();
                controllerObjects.push_back(newObject);
            }

            // Begin rendering the frame.
            // Sets the frame number to submitInfo.
            varjo_BeginFrame(session, submitInfo);

            // Get the current swap chain texture index.
            int32_t swapChainIndex = varjo_GetSwapChainCurrentIndex(session);

            // Render into the swap chain texture.
            renderer->render(frameInfo, submitInfo, swapChainIndex, donutObjects, controllerObjects);

            // Finish rendering the frame and send the texture to the compositor.
            varjo_EndFrame(session, frameInfo, submitInfo);

            controllerObjects.clear();
            // Check if we had any errors during the frame
            varjo_Error err = varjo_GetError(session);
            if (err != varjo_NoError) {
                printf("error: %s\n", varjo_GetErrorDesc(err));
                break;
            }

            lastFrameTime = frameInfo->displayTime;
            frameNumber++;

            if (enableProfiling && profiler.sampleCount() == profileFrameCount) {
                printf("Profiling finished.\n");
                break;
            }
        }
    }

    if (enableProfiling) {
        profiler.exportCSV("frame_times.csv");
    }

    openVRTracker.exit();

    // Free the frame and submit infos.
    varjo_FreeFrameInfo(frameInfo);
    varjo_FreeSubmitInfo(submitInfo);

    // Shut down the renderer
    renderer.reset();

    // We don't need the session anymore
    varjo_SessionShutDown(session);
    return 0;
}

void printHelp()
{
    printf("Usage:\n");
    printf("    --renderer=<gl|d3d11>           Renderer to be used. Defaults to d3d11.\n");
    printf("    --multi-texture                 Use one texture for each viewport instead of one big atlas texture.\n");
    printf("    --disable-animation             Disable all animation.\n");
    printf("    --profile-start-frame=<number>  Start profiling after the given frame number.\n");
    printf("    --profile-frame-count=<number>  Number of frames to profile for. Exits after all frames are profiled.\n");
    printf("    --gaze                          Use eye tracking.\n");
}

bool gotKey()
{
    HANDLE in = GetStdHandle(STD_INPUT_HANDLE);
    INPUT_RECORD input;
    DWORD count = 0;
    while (GetNumberOfConsoleInputEvents(in, &count) && count > 0) {
        if (ReadConsoleInputA(in, &input, 1, &count) && count > 0 && input.EventType == KEY_EVENT && input.Event.KeyEvent.bKeyDown) {
            if (input.Event.KeyEvent.uChar.AsciiChar == '\033') {
                printf("Quit requested.\n");
                return true;
            }
        }
    }
    return false;
}

float randomFloat(float min, float max)
{
    float value = static_cast<float>(rand()) / static_cast<float>(RAND_MAX);
    return min + (max - min) * value;
}

void createController(std::shared_ptr<IRenderer> renderer, IRenderer::Object& controllerObject)
{
    auto pentagonGeometry = GeometryGenerator::generateDonut(renderer, 3.0f, 2.0f, 5, 3);

    controllerObject.geometry = pentagonGeometry;
    controllerObject.position = glm::vec3{0, 0, 0};
    controllerObject.scale = glm::vec3{0.03, 0.03, 0.03};
    controllerObject.orientation = glm::angleAxis(glm::radians(90.0f), glm::vec3(1, 0, 0));

    printf("Created object for hand controllers\n");
}

void createGaze(std::shared_ptr<IRenderer> renderer, IRenderer::Object& gazeObject)
{
    auto circleGeometry = GeometryGenerator::generateDonut(renderer, 0.01f, 0.005f, 32, 16);

    gazeObject.geometry = circleGeometry;
    gazeObject.position = glm::vec3{0, 0, 0};
    gazeObject.scale = glm::vec3{1, 1, 1};
    gazeObject.orientation = glm::angleAxis(glm::radians(90.0f), glm::vec3(1, 0, 0));

    printf("Created object for gaze\n");
}

void createObjects(std::shared_ptr<IRenderer> renderer, std::vector<IRenderer::Object>& objects, std::vector<ObjectRotation>& rotations)
{
    auto donutGeometry = GeometryGenerator::generateDonut(renderer, 4.0f, 2.0f, 256, 64);

    const int32_t donutCount = 14;
    const int32_t rows = 5;
    const int32_t layers = 20;
    const float rowMin = -30.0f;
    const float rowMax = 30.0f;
    const float layerMin = 15.0f;
    const float layerSize = 40.0f;
    const float angle = 360.0f / static_cast<float>(donutCount);
    const float layerOffsetangle = angle * (1.0f / static_cast<float>(layers));

    srand(123);  // Keep the animations same on each run.

    for (int32_t l = 0; l < layers; ++l) {  // Layers going outward from the center.
        float offsetAngle = l * layerOffsetangle;
        float z = layerMin + (layerSize * static_cast<float>(l));

        for (int32_t r = 0; r < rows; ++r) {  // Rows going up from the bottom.
            float y = rowMin + ((rowMax - rowMin) / static_cast<float>(rows)) * static_cast<float>(r);

            for (int32_t i = 0; i < donutCount; ++i) {  // Number of donuts in a circle.
                auto rotate = glm::angleAxis(glm::radians((angle * static_cast<float>(i)) + offsetAngle), glm::vec3(0, 1, 0));

                IRenderer::Object object{};
                object.geometry = donutGeometry;
                object.position = glm::rotate(rotate, glm::vec3{0, y, z});
                object.scale = glm::vec3{1, 1, 1};
                object.orientation = rotate * glm::angleAxis(glm::radians(90.0f), glm::vec3(1, 0, 0));

                objects.push_back(object);

                // Random axis of rotation and rotation speed for each object.
                rotations.push_back({
                    glm::vec3{randomFloat(-1.0f, 1.0f), randomFloat(-1.0f, 1.0f), randomFloat(-1.0f, 1.0f)},
                    glm::radians(randomFloat(30.0f, 120.0f)),
                });
            }
        }
    }

    printf("Created %zu donuts\n", objects.size());
    printf("%zu triangles per frame\n", objects.size() * (donutGeometry->indexCount() / 3));
}
